// Bundled with Packemon: https://packemon.dev
// Platform: node, Support: stable, Format: lib
'use strict';

function ownKeys(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) { symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); } keys.push.apply(keys, symbols); } return keys; }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

Object.defineProperty(exports, '__esModule', {
  value: true
});

class ValidationError extends Error {
  constructor(message, path = '', value = undefined) {
    super(message);

    _defineProperty(this, "errors", []);

    _defineProperty(this, "path", void 0);

    _defineProperty(this, "value", void 0);

    this.name = 'ValidationError';
    this.path = path;
    this.value = value;

    if (path) {
      const key = pathKey(path);
      this.message = `${key.includes('[') ? `Invalid member "${key}".` : `Invalid field "${key}".`} ${this.message}`;
    }
  }

  addError(error) {
    const validError = error instanceof ValidationError ? error : new ValidationError(error.message);
    const hasSameError = this.errors.some(e => e.message === error.message);

    if (hasSameError) {
      return;
    }

    this.errors.push(validError);
    this.message += '\n';
    this.message += error.message.split('\n').map(line => line.match(/^\s+-/) ? `  ${line}` : `  - ${line}`).join('\n');
  }

}

function isObject(value) {
  return !!value && typeof value === 'object' && !Array.isArray(value);
}

function isSchema(value) {
  return isObject(value) && typeof value.schema === 'function' && typeof value.type === 'function' && typeof value.validate === 'function';
}

function isValidDate(value) {
  return Object.prototype.toString.call(value) === '[object Date]' && !Number.isNaN(value.getTime()) && value.toString() !== 'Invalid Date';
}

function isValidNumber(value) {
  return typeof value === 'number' && !Number.isNaN(value);
}

function isValidString(value) {
  return typeof value === 'string' && value !== '';
}

function createArray(value) {
  if (value === undefined) {
    return [];
  }

  return Array.isArray(value) ? [...value] : [value];
}

function createDate(value) {
  if (value instanceof Date) {
    return value;
  }

  if (value === undefined || value === null) {
    return new Date();
  }

  return new Date(value);
}

function createObject(value) {
  return isObject(value) ? _objectSpread({}, value) : {};
}
/**
 * Native `instanceof` checks are problematic, as cross realm checks fail.
 * They will also fail when comparing against source and compiled files.
 * So emulate an `instanceof` check by comparing constructor names.
 */


function instanceOf(object, contract) {
  if (!object || typeof object !== 'object') {
    return false;
  }

  if (object instanceof contract) {
    return true;
  }

  let current = object;

  while (current) {
    if (current.constructor.name === 'Object') {
      return false;
    }

    if (current.constructor.name === contract.name || current instanceof Error && current.name === contract.name) {
      return true;
    } // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment


    current = Object.getPrototypeOf(current);
  }

  return false;
}

function invalid(condition, message, path = '', value = undefined) {
  if (condition) {
    return;
  }

  throw new ValidationError(message, path, value);
}

function invariant(condition, message) {
  if (condition) {
    return;
  }

  throw new Error(message);
}

function logUnknown(unknownFields, pathPrefix) {
  const unknownKeys = Object.keys(unknownFields);

  if (unknownKeys.length > 0) {
    throw new Error(`${pathPrefix ? `Unknown "${pathPrefix}" fields` : 'Unknown fields'}: ${unknownKeys.join(', ')}.`);
  }
}

function pathKey(path) {
  if (path.endsWith(']')) {
    const index = path.lastIndexOf('[');
    return index > 0 ? path.slice(index) : path;
  }

  const index = path.lastIndexOf('.');
  return index > 0 ? path.slice(index + 1) : path;
}

function tryAndCollect(validator, validError, collectErrors) {
  let result = false;

  try {
    const value = validator();

    if (typeof value === 'boolean') {
      result = value;
    }
  } catch (error) {
    if (error instanceof Error && collectErrors) {
      validError.addError(error);
    } else {
      throw error;
    }
  }

  return result;
}

class OptimalError extends ValidationError {
  constructor() {
    super('');

    _defineProperty(this, "file", '');

    _defineProperty(this, "schema", '');

    this.name = 'OptimalError';
  }

  addError(error) {
    const validError = error instanceof ValidationError ? error : new ValidationError(error.message);
    this.errors.push(validError); // Avoid indenting at this level

    this.message = `${this.message}\n${error.message}`.trim();
  }

}
/**
 * Run all validation checks that have been enqueued and return a type casted value.
 * If a value is undefined, inherit the default value, else throw if required.
 * If nullable and the value is null, return early.
 */


function validate(state, validators, initialValue, path = '', {
  collectErrors = true,
  currentObject = {},
  rootObject = currentObject
} = {}) {
  const defaultValue = state.defaultValue,
        metadata = state.metadata;
  let value = initialValue; // Handle undefined

  if (value === undefined) {
    value = typeof defaultValue === 'function' ? defaultValue(path, currentObject, rootObject) : defaultValue;
    invalid(!state.required, 'Field is required and must be defined.', path);
  } else {
    if (process.env.NODE_ENV !== "production" && metadata.deprecatedMessage) {
      // eslint-disable-next-line no-console
      console.info(`Field "${path}" is deprecated. ${metadata.deprecatedMessage}`);
    }

    invalid(!state.never, 'Field should never be used.', path);
  } // Handle null


  if (value === null) {
    invalid(state.nullable, 'Null is not allowed.', path);
  } // Run validations and produce a new value


  const optimalError = new OptimalError();
  validators.forEach(test => {
    if (test.skipIfNull && value === null || test.skipIfOptional && !state.required && value === state.defaultValue) {
      return;
    }

    tryAndCollect(() => {
      const result = test.validate(value, path, {
        collectErrors,
        currentObject,
        rootObject
      });

      if (result !== undefined) {
        value = result;
      }
    }, optimalError, collectErrors);
  });

  if (optimalError.errors.length > 0) {
    throw optimalError;
  }

  return value;
}

function createSchema({
  api,
  cast,
  defaultValue,
  type
}, criteria = []) {
  const state = {
    defaultValue,
    metadata: {},
    never: false,
    nullable: false,
    required: false,
    type
  };
  const validators = [];
  criteria.forEach(crit => {
    if (typeof crit === 'function') {
      const validator = crit(state);

      if (validator) {
        validators.push(validator);
      }
    } else {
      validators.push(crit);
    }
  });
  const schema = {
    schema() {
      return type;
    },

    type() {
      return state.type;
    },

    validate(value, path, options) {
      const result = validate(state, validators, value, path, options);
      return cast && result !== null ? cast(result) : result;
    }

  };
  Object.entries(api).forEach(([name, method]) => {
    Object.defineProperty(schema, name, {
      enumerable: true,
      // Create a new schema so that our chainable API is immutable
      value: (...args) => createSchema({
        api,
        cast,
        defaultValue,
        type
      }, [...criteria, nextState => method(nextState, ...args)])
    });
  });
  return schema;
}
/**
 * Require field array to not be empty.
 */


function notEmpty$2(state, options = {}) {
  return {
    skipIfNull: true,

    validate(value, path) {
      var _options$message;

      invalid(value.length > 0, (_options$message = options.message) !== null && _options$message !== void 0 ? _options$message : 'Array cannot be empty.', path, value);
    }

  };
}
/**
 * Require field array items to be of a specific schema type.
 * Will rebuild the array and type cast values.
 */


function of$5(state, itemsSchema) {
  if (process.env.NODE_ENV !== "production") {
    invariant(isSchema(itemsSchema), 'A schema is required for array items.');
  }

  state.type += `<${itemsSchema.type()}>`;
  return {
    skipIfNull: true,

    validate(value, path, validateOptions) {
      if (!Array.isArray(value)) {
        return [];
      }

      const nextValue = [...value];
      value.forEach((item, i) => {
        nextValue[i] = itemsSchema.validate(item, `${path}[${i}]`, validateOptions);
      });
      return nextValue;
    }

  };
}
/**
 * Require field array to be of a specific size.
 */


function sizeOf$2(state, size, options = {}) {
  if (process.env.NODE_ENV !== "production") {
    invariant(typeof size === 'number' && size > 0, 'Size of requires a non-zero positive number.');
  }

  return {
    skipIfNull: true,

    validate(value, path) {
      var _options$message2;

      invalid(value.length === size, (_options$message2 = options.message) !== null && _options$message2 !== void 0 ? _options$message2 : `Array length must be ${size}.`, path, value);
    }

  };
}

const arrayCriteria = /*#__PURE__*/Object.freeze({
  __proto__: null,
  notEmpty: notEmpty$2,
  of: of$5,
  sizeOf: sizeOf$2
});
/**
 * Require this field to only be false.
 */

function onlyFalse(state, options = {}) {
  state.defaultValue = false;
  return {
    validate(value, path) {
      var _options$message3;

      invalid(!value, (_options$message3 = options.message) !== null && _options$message3 !== void 0 ? _options$message3 : 'May only be `false`.', path, value);
    }

  };
}
/**
 * Require this field to only be true.
 */


function onlyTrue(state, options = {}) {
  state.defaultValue = true;
  return {
    validate(value, path) {
      var _options$message4;

      invalid(value, (_options$message4 = options.message) !== null && _options$message4 !== void 0 ? _options$message4 : 'May only be `true`.', path, value);
    }

  };
}

const booleanCriteria = /*#__PURE__*/Object.freeze({
  __proto__: null,
  onlyFalse: onlyFalse,
  onlyTrue: onlyTrue
});
/**
 * Require this field to be an instance of the defined class.
 */

function of$4(state, ref, loose = false) {
  var _ref$name;

  if (process.env.NODE_ENV !== "production") {
    invariant(typeof ref === 'function', 'A class reference is required.');
  }

  state.type = (_ref$name = ref.name) !== null && _ref$name !== void 0 ? _ref$name : ref.constructor.name;
  return {
    skipIfNull: true,

    validate(value, path) {
      invalid(typeof ref === 'function' && (value instanceof ref || loose && isObject(value) && instanceOf(value, ref)), `Must be an instance of "${state.type}".`, path, value);
    }

  };
}

const classCriteria = /*#__PURE__*/Object.freeze({
  __proto__: null,
  of: of$4
});
/**
 * Map a list of field names that must be defined alongside this field.
 */

function and(state, ...keys) {
  if (process.env.NODE_ENV !== "production") {
    invariant(keys.length > 0, 'AND requires a list of field names.');
  }

  return {
    validate(value, path, {
      currentObject
    }) {
      const andKeys = [...new Set([pathKey(path), ...keys])].sort();
      const undefs = andKeys.filter(key => (currentObject === null || currentObject === void 0 ? void 0 : currentObject[key]) === undefined || (currentObject === null || currentObject === void 0 ? void 0 : currentObject[key]) === null); // Only error once when one of the struct is defined

      if (undefs.length === andKeys.length) {
        return;
      }

      invalid(undefs.length === 0, `All of these fields must be defined: ${andKeys.join(', ')}`);
    }

  };
}
/**
 * Set a callback to run custom validation logic.
 */


function custom$1(state, validator) {
  if (process.env.NODE_ENV !== "production") {
    invariant(typeof validator === 'function', 'Custom requires a validation function.');
  }

  return {
    validate(value, path, validateOptions) {
      try {
        validator(value, path, validateOptions);
      } catch (error) {
        if (error instanceof Error) {
          invalid(false, error.message, path, value);
        }
      }
    }

  };
}
/**
 * Set a message to log when this field is present.
 */


function deprecate(state, message) {
  if (process.env.NODE_ENV !== "production") {
    invariant(isValidString(message), 'A non-empty string is required for deprecated messages.');
  }

  state.metadata.deprecatedMessage = message;
}
/**
 * Mark that this field should never be used.
 */


function never(state) {
  state.defaultValue = undefined;
  state.never = true;
}
/**
 * Disallow null values.
 */


function notNullable(state) {
  state.nullable = false;
}
/**
 * Require this field to NOT be explicitly defined.
 */


function notRequired(state) {
  state.required = false;
}
/**
 * Allow null values.
 */


function nullable(state) {
  state.nullable = true;
}
/**
 * Mark that this field can ONLY use a value that matches the default value.
 */


function only(state) {
  const defaultValue = state.defaultValue;

  if (process.env.NODE_ENV !== "production") {
    invariant(defaultValue !== null && defaultValue !== undefined, 'Only requires a non-empty default value.');
  }

  return {
    validate(value, path) {
      invalid(value === defaultValue, `Value may only be "${defaultValue}".`, path, value);
    }

  };
}
/**
 * Map a list of field names that must have at least 1 defined.
 */


function or(state, ...keys) {
  if (process.env.NODE_ENV !== "production") {
    invariant(keys.length > 0, 'OR requires a list of field names.');
  }

  return {
    validate(value, path, {
      currentObject
    }) {
      const orKeys = [...new Set([pathKey(path), ...keys])].sort();
      const defs = orKeys.filter(key => (currentObject === null || currentObject === void 0 ? void 0 : currentObject[key]) !== undefined && (currentObject === null || currentObject === void 0 ? void 0 : currentObject[key]) !== null);
      invalid(defs.length > 0, `At least one of these fields must be defined: ${orKeys.join(', ')}`);
    }

  };
}
/**
 * Require this field to be explicitly defined.
 */


function required(state) {
  state.required = true;
}
/**
 * Validate with a specific schema when a condition is met.
 */


function when(state, condition, pass, fail) {
  if (process.env.NODE_ENV !== "production") {
    invariant(isSchema(pass), 'A schema is required when the condition passes.');
  }

  if (fail !== undefined) {
    if (process.env.NODE_ENV !== "production") {
      invariant(isSchema(fail), 'A schema is required when the condition fails.');
    }
  }

  return {
    validate(value, path, validateOptions) {
      const passed = typeof condition === 'function' ? condition(value, validateOptions.currentObject, validateOptions.rootObject) : condition === value;

      if (passed) {
        return pass.validate(value, path);
      }

      if (fail) {
        return fail.validate(value, path, validateOptions);
      }

      return undefined;
    }

  };
}
/**
 * Map a list of field names that must not be defined alongside this field.
 */


function xor(state, ...keys) {
  if (process.env.NODE_ENV !== "production") {
    invariant(keys.length > 0, 'XOR requires a list of field names.');
  }

  return {
    validate(value, path, {
      currentObject
    }) {
      const xorKeys = [...new Set([pathKey(path), ...keys])].sort();
      const defs = xorKeys.filter(key => (currentObject === null || currentObject === void 0 ? void 0 : currentObject[key]) !== undefined && (currentObject === null || currentObject === void 0 ? void 0 : currentObject[key]) !== null);
      invalid(defs.length === 1, `Only one of these fields may be defined: ${xorKeys.join(', ')}`);
    }

  };
}

const commonCriteria = /*#__PURE__*/Object.freeze({
  __proto__: null,
  and: and,
  custom: custom$1,
  deprecate: deprecate,
  never: never,
  notNullable: notNullable,
  notRequired: notRequired,
  nullable: nullable,
  only: only,
  or: or,
  required: required,
  when: when,
  xor: xor
});
/**
 * Require field value to be after the provided date.
 */

function after(state, date, options = {}) {
  const afterDate = createDate(date);

  if (process.env.NODE_ENV !== "production") {
    invariant(isValidDate(afterDate), 'After date must be a valid date.');
  }

  return {
    skipIfNull: true,

    validate(value, path) {
      var _options$message5;

      invalid(isValidDate(value) && value > afterDate, (_options$message5 = options.message) !== null && _options$message5 !== void 0 ? _options$message5 : `Date must come after ${afterDate.toLocaleDateString()}.`, path, value);
    }

  };
}
/**
 * Require field value to be before the provided date.
 */


function before(state, date, options = {}) {
  const beforeDate = createDate(date);

  if (process.env.NODE_ENV !== "production") {
    invariant(isValidDate(beforeDate), 'Before date must be a valid date.');
  }

  return {
    skipIfNull: true,

    validate(value, path) {
      var _options$message6;

      invalid(isValidDate(value) && value < beforeDate, (_options$message6 = options.message) !== null && _options$message6 !== void 0 ? _options$message6 : `Date must come before ${beforeDate.toLocaleDateString()}.`, path, value);
    }

  };
}
/**
 * Require field value to be between 2 date ranges.
 */


function between$1(state, start, end, options = {}) {
  const startDate = createDate(start);
  const endDate = createDate(end);

  if (process.env.NODE_ENV !== "production") {
    invariant(isValidDate(startDate), 'Between start date must be a valid date.');
  }

  if (process.env.NODE_ENV !== "production") {
    invariant(isValidDate(endDate), 'Between end date must be a valid date.');
  }

  return {
    skipIfNull: true,

    validate(value, path) {
      var _options$message7;

      invalid(isValidDate(value) && (options.inclusive ? value >= startDate && value <= endDate : value > startDate && value < endDate), (_options$message7 = options.message) !== null && _options$message7 !== void 0 ? _options$message7 : `Date must be between ${startDate.toLocaleDateString()} and ${endDate.toLocaleDateString()}${options.inclusive ? ' inclusive' : ''}.`, path, value);
    }

  };
}

const dateCriteria = /*#__PURE__*/Object.freeze({
  __proto__: null,
  after: after,
  before: before,
  between: between$1
});
/**
 * Require field value to be between 2 numbers.
 */

function between(state, min, max, options = {}) {
  if (process.env.NODE_ENV !== "production") {
    invariant(isValidNumber(min) && isValidNumber(max), 'Between requires a minimum and maximum number.');
  }

  return {
    skipIfNull: true,

    validate(value, path) {
      var _options$message8;

      invalid(isValidNumber(value) && (options.inclusive ? value >= min && value <= max : value > min && value < max), (_options$message8 = options.message) !== null && _options$message8 !== void 0 ? _options$message8 : `Number must be between ${min} and ${max}${options.inclusive ? ' inclusive' : ''}.`, path, value);
    }

  };
}
/**
 * Require field value to be a float (includes a decimal).
 */


function float(state, options = {}) {
  return {
    skipIfNull: true,

    validate(value, path) {
      var _options$message9;

      invalid(isValidNumber(value) && value % 1 !== 0, (_options$message9 = options.message) !== null && _options$message9 !== void 0 ? _options$message9 : 'Number must be a float.', path, value);
    }

  };
}
/**
 * Require field value to be greater than a number.
 */


function gt(state, min, options = {}) {
  if (process.env.NODE_ENV !== "production") {
    invariant(isValidNumber(min), 'Greater-than requires a minimum number.');
  }

  return {
    skipIfNull: true,

    validate(value, path) {
      if (options.inclusive) {
        var _options$message10;

        invalid(isValidNumber(value) && value >= min, (_options$message10 = options.message) !== null && _options$message10 !== void 0 ? _options$message10 : `Number must be greater than or equal to ${min}.`, path, value);
      } else {
        var _options$message11;

        invalid(isValidNumber(value) && value > min, (_options$message11 = options.message) !== null && _options$message11 !== void 0 ? _options$message11 : `Number must be greater than ${min}.`, path, value);
      }
    }

  };
}
/**
 * Require field value to be greater than or equals to a number.
 */


function gte(state, min, options = {}) {
  return gt(state, min, _objectSpread(_objectSpread({}, options), {}, {
    inclusive: true
  }));
}
/**
 * Require field value to be an integer.
 */


function int(state, options = {}) {
  return {
    skipIfNull: true,

    validate(value, path) {
      var _options$message12;

      invalid(Number.isSafeInteger(value), (_options$message12 = options.message) !== null && _options$message12 !== void 0 ? _options$message12 : 'Number must be an integer.', path, value);
    }

  };
}
/**
 * Require field value to be less than a number.
 */


function lt(state, max, options = {}) {
  if (process.env.NODE_ENV !== "production") {
    invariant(isValidNumber(max), 'Less-than requires a maximum number.');
  }

  return {
    skipIfNull: true,

    validate(value, path) {
      if (options.inclusive) {
        var _options$message13;

        invalid(isValidNumber(value) && value <= max, (_options$message13 = options.message) !== null && _options$message13 !== void 0 ? _options$message13 : `Number must be less than or equal to ${max}.`, path, value);
      } else {
        var _options$message14;

        invalid(isValidNumber(value) && value < max, (_options$message14 = options.message) !== null && _options$message14 !== void 0 ? _options$message14 : `Number must be less than ${max}.`, path, value);
      }
    }

  };
}
/**
 * Require field value to be less than or equals to a number.
 */


function lte(state, max, options = {}) {
  return lt(state, max, _objectSpread(_objectSpread({}, options), {}, {
    inclusive: true
  }));
}
/**
 * Require field value to be negative and not zero.
 */


function negative(state, options = {}) {
  return {
    skipIfNull: true,

    validate(value, path) {
      var _options$message15;

      invalid(isValidNumber(value) && value < 0, (_options$message15 = options.message) !== null && _options$message15 !== void 0 ? _options$message15 : 'Number must be negative.', path, value);
    }

  };
}
/**
 * Require field value to be one of the provided numbers.
 */


function oneOf$1(state, list, options = {}) {
  if (process.env.NODE_ENV !== "production") {
    invariant(Array.isArray(list) && list.length > 0 && list.every(item => isValidNumber(item)), 'One of requires an array of numbers.');
  }

  return {
    skipIfNull: true,

    validate(value, path) {
      var _options$message16;

      invalid(list.includes(value), (_options$message16 = options.message) !== null && _options$message16 !== void 0 ? _options$message16 : `Number must be one of: ${list.join(', ')}`, path, value);
    }

  };
}
/**
 * Require field value to be positive and not zero.
 */


function positive(state, options = {}) {
  return {
    skipIfNull: true,

    validate(value, path) {
      var _options$message17;

      invalid(isValidNumber(value) && value > 0, (_options$message17 = options.message) !== null && _options$message17 !== void 0 ? _options$message17 : 'Number must be positive.', path, value);
    }

  };
}

const numberCriteria = /*#__PURE__*/Object.freeze({
  __proto__: null,
  between: between,
  float: float,
  gt: gt,
  gte: gte,
  int: int,
  lt: lt,
  lte: lte,
  negative: negative,
  oneOf: oneOf$1,
  positive: positive
});
/**
 * Require field object keys to be of a string schema type.
 */

function keysOf(state, keysSchema, options = {}) {
  if (process.env.NODE_ENV !== "production") {
    invariant(isSchema(keysSchema) && keysSchema.schema() === 'string', 'A string schema is required for object keys.');
  }

  return {
    skipIfNull: true,

    validate(value, path, validateOptions) {
      if (isObject(value)) {
        Object.keys(value).forEach(key => {
          try {
            // Dont pass path so its not included in the error message
            keysSchema.validate(key, '', validateOptions);
          } catch (error) {
            if (error instanceof Error) {
              var _options$message18;

              invalid(false, `Invalid key "${key}". ${(_options$message18 = options.message) !== null && _options$message18 !== void 0 ? _options$message18 : error.message}`, path, value);
            }
          }
        });
      }
    }

  };
}
/**
 * Require field object to not be empty.
 */


function notEmpty$1(state, options = {}) {
  return {
    skipIfNull: true,

    validate(value, path) {
      var _options$message19;

      invalid(Object.keys(value).length > 0, (_options$message19 = options.message) !== null && _options$message19 !== void 0 ? _options$message19 : 'Object cannot be empty.', path, value);
    }

  };
}
/**
 * Require field object values to be of a specific schema type.
 * Will rebuild the object and type cast values.
 */


function of$3(state, valuesSchema) {
  if (process.env.NODE_ENV !== "production") {
    invariant(isSchema(valuesSchema), 'A schema is required for object values.');
  }

  state.type += `<${valuesSchema.type()}>`;
  return {
    skipIfNull: true,

    validate(value, path, validateOptions) {
      if (!isObject(value)) {
        return {};
      }

      const nextValue = _objectSpread({}, value);

      Object.keys(value).forEach(baseKey => {
        const key = baseKey;
        nextValue[key] = valuesSchema.validate(value[key], path ? `${path}.${key}` : String(key), validateOptions);
      });
      return nextValue;
    }

  };
}
/**
 * Require field object to be of a specific size.
 */


function sizeOf$1(state, size, options = {}) {
  if (process.env.NODE_ENV !== "production") {
    invariant(typeof size === 'number' && size > 0, 'Size of requires a non-zero positive number.');
  }

  return {
    skipIfNull: true,

    validate(value, path) {
      var _options$message20;

      invalid(Object.keys(value).length === size, (_options$message20 = options.message) !== null && _options$message20 !== void 0 ? _options$message20 : size === 1 ? `Object must have ${size} property.` : `Object must have ${size} properties.`, path, value);
    }

  };
}

const objectCriteria = /*#__PURE__*/Object.freeze({
  __proto__: null,
  keysOf: keysOf,
  notEmpty: notEmpty$1,
  of: of$3,
  sizeOf: sizeOf$1
});
/**
 * Require a shape to be an exact shape.
 * No more and no less of the same properties.
 */

function exact(state, value = true) {
  state.metadata.exact = value;
}
/**
 * Require field to be an object with every property being a schema type.
 * Will rebuild the object (if not a class instance) and type cast values.
 */


function of$2(state, schemas) {
  if (process.env.NODE_ENV !== "production") {
    invariant(isObject(schemas) && Object.keys(schemas).length > 0 && Object.values(schemas).every(isSchema), 'A non-empty object of schemas are required for a shape.');
  }

  const types = Object.entries(schemas).map(([key, value]) => `${key}: ${value.type()}`);
  state.type += `<{ ${types.join(', ')} }>`;
  return {
    skipIfNull: true,

    validate(value, path, validateOptions) {
      if (value) {
        invalid(isObject(value), 'Value passed to shape must be an object.', path, value);
      }

      const isPlainObject = value.constructor === Object;
      const unknown = isPlainObject ? _objectSpread({}, value) : {};
      const shape = {};
      const collectionError = new ValidationError('The following validations have failed:', path, value);
      Object.keys(schemas).forEach(prop => {
        const key = prop;
        const schema = schemas[key];
        tryAndCollect(() => {
          shape[key] = schema.validate(value[key], path ? `${path}.${key}` : String(key), _objectSpread(_objectSpread({}, validateOptions), {}, {
            currentObject: value
          }));
        }, collectionError, validateOptions.collectErrors); // Delete the prop and mark it as known

        delete unknown[key];
      });

      if (collectionError.errors.length > 0) {
        throw collectionError;
      } // Handle unknown fields


      if (state.metadata.exact) {
        if (process.env.NODE_ENV !== "production") {
          logUnknown(unknown, path);
        }
      } else {
        Object.assign(shape, unknown);
      } // Dont replace class instance with plain objects


      return isPlainObject ? shape : value;
    }

  };
}

const shapeCriteria = /*#__PURE__*/Object.freeze({
  __proto__: null,
  exact: exact,
  of: of$2
});
/**
 * Require field value to contain a provided string.
 */

function contains(state, token, options = {}) {
  if (process.env.NODE_ENV !== "production") {
    invariant(isValidString(token), 'Contains requires a non-empty token.');
  }

  return {
    skipIfNull: true,
    skipIfOptional: true,

    validate(value, path) {
      var _options$index, _options$message21;

      invalid(value.includes(token, (_options$index = options.index) !== null && _options$index !== void 0 ? _options$index : 0), (_options$message21 = options.message) !== null && _options$message21 !== void 0 ? _options$message21 : `String does not include "${token}".`, path, value);
    }

  };
}
/**
 * Require field value to match a defined regex pattern.
 */


function match(state, pattern, options = {}) {
  if (process.env.NODE_ENV !== "production") {
    invariant(pattern instanceof RegExp, 'Match requires a regular expression to match against.');
  }

  return {
    skipIfNull: true,
    skipIfOptional: true,

    validate(value, path) {
      var _options$message22;

      invalid(!!value.match(pattern), `${(_options$message22 = options.message) !== null && _options$message22 !== void 0 ? _options$message22 : 'String does not match.'} (pattern "${pattern.source}")`, path, value);
    }

  };
}
/**
 * Require field value to be formatted in camel case (fooBar).
 */


function camelCase(state, options = {}) {
  return match(state, /^[a-z][a-zA-Z0-9]+$/u, _objectSpread({
    message: 'String must be in camel case.'
  }, options));
}
/**
 * Require field value to be formatted in kebab case (foo-bar).
 */


function kebabCase(state, options = {}) {
  return match(state, /^[a-z][a-z0-9-]+$/u, _objectSpread({
    message: 'String must be in kebab case.'
  }, options));
}
/**
 * Require field value to be formatted in pascal case (FooBar).
 */


function pascalCase(state, options = {}) {
  return match(state, /^[A-Z][a-zA-Z0-9]+$/u, _objectSpread({
    message: 'String must be in pascal case.'
  }, options));
}
/**
 * Require field value to be formatted in snake case (foo_bar).
 */


function snakeCase(state, options = {}) {
  return match(state, /^[a-z][a-z0-9_]+$/u, _objectSpread({
    message: 'String must be in snake case.'
  }, options));
}
/**
 * Require field value to not be an empty string.
 */


function notEmpty(state, options = {}) {
  return {
    skipIfNull: true,

    validate(value, path) {
      var _options$message23;

      invalid(isValidString(value), (_options$message23 = options.message) !== null && _options$message23 !== void 0 ? _options$message23 : 'String cannot be empty.', path, value);
    }

  };
}
/**
 * Require field value to be one of the provided string.
 */


function oneOf(state, list, options = {}) {
  if (process.env.NODE_ENV !== "production") {
    invariant(Array.isArray(list) && list.length > 0 && list.every(item => typeof item === 'string'), 'One of requires an array of strings.');
  }

  return {
    skipIfNull: true,

    validate(value, path) {
      var _options$message24;

      invalid(list.includes(value), (_options$message24 = options.message) !== null && _options$message24 !== void 0 ? _options$message24 : `String must be one of: ${list.join(', ')}`, path, value);
    }

  };
}
/**
 * Require field value to be all lower case.
 */


function lowerCase(state, options = {}) {
  return {
    skipIfNull: true,

    validate(value, path) {
      var _options$message25;

      invalid(value === value.toLocaleLowerCase(), (_options$message25 = options.message) !== null && _options$message25 !== void 0 ? _options$message25 : 'String must be lower cased.', path, value);
    }

  };
}
/**
 * Require field value to be all upper case.
 */


function upperCase(state, options = {}) {
  return {
    skipIfNull: true,

    validate(value, path) {
      var _options$message26;

      invalid(value === value.toLocaleUpperCase(), (_options$message26 = options.message) !== null && _options$message26 !== void 0 ? _options$message26 : 'String must be upper cased.', path, value);
    }

  };
}
/**
 * Require field array to be of a specific size.
 */


function sizeOf(state, size, options = {}) {
  if (process.env.NODE_ENV !== "production") {
    invariant(typeof size === 'number' && size > 0, 'Size requires a non-zero positive number.');
  }

  return {
    skipIfNull: true,

    validate(value, path) {
      var _options$message27;

      invalid(value.length === size, (_options$message27 = options.message) !== null && _options$message27 !== void 0 ? _options$message27 : `String length must be ${size}.`, path, value);
    }

  };
}

const stringCriteria = /*#__PURE__*/Object.freeze({
  __proto__: null,
  contains: contains,
  match: match,
  camelCase: camelCase,
  kebabCase: kebabCase,
  pascalCase: pascalCase,
  snakeCase: snakeCase,
  notEmpty: notEmpty,
  oneOf: oneOf,
  lowerCase: lowerCase,
  upperCase: upperCase,
  sizeOf: sizeOf
});
/**
 * Require field array items to be of a specific schema type.
 * Will rebuild the array and type cast values.
 */

function of$1(state, itemsSchemas) {
  if (process.env.NODE_ENV !== "production") {
    invariant(Array.isArray(itemsSchemas) && itemsSchemas.length > 0 && itemsSchemas.every(isSchema), 'A non-empty array of schemas are required for a tuple.');
  }

  state.type = `tuple<${itemsSchemas.map(item => item.type()).join(', ')}>`;
  return {
    skipIfNull: true,

    validate(value, path, validateOptions) {
      invalid(Array.isArray(value) && value.length <= itemsSchemas.length, `Value must be a tuple with ${itemsSchemas.length} items.`, path, value);
      return itemsSchemas.map((item, i) => item.validate(value[i], `${path}[${i}]`, validateOptions));
    }

  };
}

const tupleCriteria = /*#__PURE__*/Object.freeze({
  __proto__: null,
  of: of$1
});

function typeOf(value) {
  if (Array.isArray(value)) {
    return 'array/tuple';
  }

  if (isObject(value)) {
    return value.constructor === Object ? 'object/shape' : 'class';
  }

  return typeof value;
}
/**
 * Require field value to be one of a specific schema type.
 */


function of(state, schemas) {
  if (process.env.NODE_ENV !== "production") {
    invariant(Array.isArray(schemas) && schemas.length > 0 && schemas.every(isSchema), 'A non-empty array of schemas are required for a union.');
  }

  state.type = schemas.map(item => item.type()).join(' | ');
  return {
    skipIfNull: true,

    validate(value, path, validateOptions) {
      let nextValue = value;
      const allowedValues = schemas.map(schema => schema.type()).join(', ');
      const valueType = typeOf(value);
      const collectionError = new ValidationError(`Received ${valueType} with the following failures:`, path, value);
      const passed = schemas.some(schema => {
        const schemaType = schema.schema();

        if (schemaType === 'union') {
          invalid(false, 'Nested unions are not supported.', path);
        }

        return tryAndCollect(() => {
          if (valueType === schemaType || valueType === 'object/shape' && schemaType === 'object' || valueType === 'object/shape' && schemaType === 'shape' || valueType === 'array/tuple' && schemaType === 'array' || valueType === 'array/tuple' && schemaType === 'tuple' || schemaType === 'custom') {
            // Dont pass path so its not included in the error message
            nextValue = schema.validate(value, '', validateOptions);
            return true;
          }

          return false;
        }, collectionError, validateOptions.collectErrors);
      });

      if (!passed) {
        if (collectionError.errors.length > 0) {
          throw collectionError;
        } else {
          invalid(false, `Value must be one of: ${allowedValues}.`, path, value);
        }
      }

      return nextValue;
    }

  };
}

const unionCriteria = /*#__PURE__*/Object.freeze({
  __proto__: null,
  of: of
});

function array(defaultValue = []) {
  return createSchema({
    api: _objectSpread(_objectSpread({}, commonCriteria), arrayCriteria),
    cast: createArray,
    defaultValue,
    type: 'array'
  }, [{
    skipIfNull: true,

    validate(value, path) {
      invalid(Array.isArray(value), 'Must be an array.', path, value);
    }

  }]);
}

function object(defaultValue) {
  return createSchema({
    api: _objectSpread(_objectSpread({}, commonCriteria), objectCriteria),
    cast: createObject,
    defaultValue: defaultValue !== null && defaultValue !== void 0 ? defaultValue : {},
    type: 'object'
  }, [{
    skipIfNull: true,

    validate(value, path) {
      invalid(isObject(value), 'Must be a plain object.', path, value);
    }

  }]);
} // eslint-disable-next-line @typescript-eslint/no-explicit-any


function func(defaultValue) {
  return createSchema({
    api: _objectSpread({}, commonCriteria),
    defaultValue,
    type: 'function'
  }, [{
    skipIfNull: true,
    skipIfOptional: true,

    validate(value, path) {
      invalid(typeof value === 'function', 'Must be a function.', path, value);
    }

  }]);
} // This is similar to shape, but we want to control the validation


function schema() {
  const shape = createSchema({
    api: _objectSpread(_objectSpread({}, commonCriteria), shapeCriteria),
    cast: createObject,
    type: 'shape'
  }, [{
    skipIfNull: true,

    validate(value, path) {
      invalid(isObject(value), 'Must be a schema.', path, value);
    }

  }]);
  return shape.of({
    schema: func().notNullable().required(),
    type: func().notNullable().required(),
    validate: func().notNullable().required()
  });
}

function blueprint(defaultValue)
/* infer */
{
  return object(defaultValue).of(schema().notNullable());
}

function bool(defaultValue = false) {
  return createSchema({
    api: _objectSpread(_objectSpread({}, commonCriteria), booleanCriteria),
    cast: Boolean,
    defaultValue,
    type: 'boolean'
  }, [{
    skipIfNull: true,

    validate(value, path) {
      invalid(typeof value === 'boolean', 'Must be a boolean.', path, value);
    }

  }]);
}

function custom(validator, defaultValue) {
  return createSchema({
    api: _objectSpread({}, commonCriteria),
    defaultValue,
    type: 'custom'
  }).custom(validator);
}

function date(defaultValue) {
  return createSchema({
    api: _objectSpread(_objectSpread({}, commonCriteria), dateCriteria),
    cast: createDate,
    defaultValue: defaultValue !== null && defaultValue !== void 0 ? defaultValue : new Date(),
    type: 'date'
  }, [{
    skipIfNull: true,

    validate(value, path) {
      const time = createDate(value);
      invalid(isValidDate(time), 'Must be a string, number, or `Date` that resolves to a valid date.', path, value);
      return time;
    }

  }]);
}

function instance() {
  return createSchema({
    api: _objectSpread(_objectSpread({}, commonCriteria), classCriteria),
    defaultValue: null,
    type: 'class'
  }, [state => ({
    skipIfNull: true,

    validate(value, path) {
      invalid(isObject(value) && value.constructor !== Object, state.type === 'class' ? 'Must be a class instance.' : `Must be an instance of ${state.type}.`, path, value);
    }

  })]).nullable();
}

function lazy(factory, defaultValue) {
  if (process.env.NODE_ENV !== "production") {
    invariant(typeof factory === 'function', 'Lazy requires a schema factory function.');
  }

  return createSchema({
    api: _objectSpread({}, commonCriteria),
    defaultValue,
    type: 'lazy'
  }, [{
    // Avoid recursion by returning early and using the provided default value
    skipIfNull: true,
    skipIfOptional: true,

    validate(value, path, validateOptions) {
      const schema = factory(value);

      if (process.env.NODE_ENV !== "production") {
        invariant(isSchema(schema), 'Factory must return a schema.');
      }

      return schema.validate(value, path, validateOptions);
    }

  }]);
}

function cast$1(value) {
  return value === undefined ? 0 : Number(value);
}

function number(defaultValue = 0) {
  return createSchema({
    api: _objectSpread(_objectSpread({}, commonCriteria), numberCriteria),
    cast: cast$1,
    defaultValue,
    type: 'number'
  }, [{
    skipIfNull: true,

    validate(value, path) {
      invalid(typeof value === 'number', 'Must be a number.', path, value);
    }

  }]);
}

function regex()
/* infer */
{
  return instance().of(RegExp);
}

function shape(blueprint) {
  return createSchema({
    api: _objectSpread(_objectSpread({}, commonCriteria), shapeCriteria),
    cast: createObject,
    type: 'shape'
  }, [{
    skipIfNull: true,

    validate(value, path) {
      if (value === undefined) {
        // Will be built from its items
        return {};
      }

      invalid(isObject(value), 'Must be a shaped object.', path, value);
      return value;
    }

  }]).of(blueprint);
}

function cast(value) {
  return value === undefined ? '' : String(value);
}

function string(defaultValue = '') {
  return createSchema({
    api: _objectSpread(_objectSpread({}, commonCriteria), stringCriteria),
    cast,
    defaultValue,
    type: 'string'
  }, [{
    skipIfNull: true,

    validate(value, path) {
      invalid(typeof value === 'string', 'Must be a string.', path, value);
    }

  }]);
}

function tuple(schemas) {
  return createSchema({
    api: _objectSpread(_objectSpread({}, commonCriteria), tupleCriteria),
    // @ts-expect-error Ignore this, it's safe
    cast: createArray,
    type: 'tuple'
  }, [{
    skipIfNull: true,

    validate(value, path) {
      if (value === undefined) {
        // Will be built from its items
        return [];
      }

      invalid(Array.isArray(value), 'Must be a tuple.', path, value);
      return value;
    }

  }]).of(schemas);
}

function union(defaultValue) {
  return createSchema({
    api: _objectSpread(_objectSpread({}, commonCriteria), unionCriteria),
    defaultValue,
    type: 'union'
  });
}

const index = /*#__PURE__*/Object.freeze({
  __proto__: null,
  array: array,
  blueprint: blueprint,
  bool: bool,
  custom: custom,
  date: date,
  func: func,
  instance: instance,
  lazy: lazy,
  number: number,
  object: object,
  regex: regex,
  schema: schema,
  shape: shape,
  string: string,
  tuple: tuple,
  union: union
});

function createPredicate(schema) {
  return value => {
    try {
      schema.validate(value);
    } catch {
      return false;
    }

    return true;
  };
}

function optimal(blueprint, baseOpts = {}) {
  const options = {};
  let schema = shape(blueprint);

  function configure(nextOpts) {
    if (!isObject(nextOpts)) {
      throw new TypeError('Optimal options must be a plain object.');
    }

    Object.assign(options, nextOpts);
    schema = schema.exact(!options.unknown);
  }

  configure(baseOpts);
  return {
    configure,

    validate(struct, validateOptions) {
      const object = struct;

      try {
        var _options$prefix;

        return schema.validate(struct, (_options$prefix = options.prefix) !== null && _options$prefix !== void 0 ? _options$prefix : '', _objectSpread(_objectSpread({
          collectErrors: true
        }, validateOptions), {}, {
          currentObject: object,
          rootObject: object
        }));
      } catch (error) {
        let invalid;

        if (error instanceof OptimalError) {
          invalid = error;
        } else {
          invalid = new OptimalError();
          invalid.addError(error);
        }

        if (options.name) {
          invalid.schema = options.name;
        }

        if (options.file) {
          invalid.file = options.file;
        }

        throw invalid;
      }
    }

  };
}

exports.OptimalError = OptimalError;
exports.ValidationError = ValidationError;
exports.array = array;
exports.arrayCriteria = arrayCriteria;
exports.blueprint = blueprint;
exports.bool = bool;
exports.booleanCriteria = booleanCriteria;
exports.classCriteria = classCriteria;
exports.commonCriteria = commonCriteria;
exports.createPredicate = createPredicate;
exports.createSchema = createSchema;
exports.custom = custom;
exports.date = date;
exports.dateCriteria = dateCriteria;
exports.func = func;
exports.instance = instance;
exports.lazy = lazy;
exports.number = number;
exports.numberCriteria = numberCriteria;
exports.object = object;
exports.objectCriteria = objectCriteria;
exports.optimal = optimal;
exports.regex = regex;
exports.schema = schema;
exports.schemas = index;
exports.shape = shape;
exports.shapeCriteria = shapeCriteria;
exports.string = string;
exports.stringCriteria = stringCriteria;
exports.tuple = tuple;
exports.tupleCriteria = tupleCriteria;
exports.union = union;
exports.unionCriteria = unionCriteria;
//# sourceMappingURL=index.js.map
